# EpyNN/epynn/lstm/forward.py
# Related third party imports
import numpy as np


def initialize_forward(layer, A):
    """Forward cache initialization.

    :param layer: An instance of LSTM layer.
    :type layer: :class:`epynn.lstm.models.LSTM`

    :param A: Output of forward propagation from previous layer.
    :type A: :class:`numpy.ndarray`

    :return: Input of forward propagation for current layer.
    :rtype: :class:`numpy.ndarray`

    :return: Previous hidden state initialized with zeros.
    :rtype: :class:`numpy.ndarray`

    :return: Previous memory state initialized with zeros.
    :rtype: :class:`numpy.ndarray`
    """
    X = layer.fc['X'] = A

    cache_keys = ['h', 'hp', 'o_', 'o', 'i_', 'i', 'f_', 'f', 'g_', 'g', 'C_', 'Cp_', 'C']
    layer.fc.update({k: np.zeros(layer.fs['h']) for k in cache_keys})

    h = layer.fc['h'][:, 0]    # Hidden state
    C_ = layer.fc['C_'][:, 0]  # Memory state

    return X, h, C_


def lstm_forward(layer, A):
    """Forward propagate signal to next layer.
    """
    # (1) Initialize cache, hidden and memory states
    X, h, C_ = initialize_forward(layer, A)

    # Iterate over sequence steps
    for s in range(layer.d['s']):

        # (2s) Slice sequence (m, s, e) w.r.t to step
        X = layer.fc['X'][:, s]

        # (3s) Retrieve previous states
        hp = layer.fc['hp'][:, s] = h       # (3.1s) Hidden
        Cp_ = layer.fc['Cp_'][:, s] = C_    # (3.2s) Memory

        # (4s) Activate forget gate
        f_ = layer.fc['f_'][:, s] = (
            np.dot(X, layer.p['Uf'])
            + np.dot(hp, layer.p['Vf'])
            + layer.p['bf']
        )   # (4.1s)

        f = layer.fc['f'][:, s] = layer.activate_forget(f_)      # (4.2s)

        # (5s) Activate input gate
        i_ = layer.fc['i_'][:, s] = (
            np.dot(X, layer.p['Ui'])
            + np.dot(hp, layer.p['Vi'])
            + layer.p['bi']
        )   # (5.1s)

        i = layer.fc['i'][:, s] = layer.activate_input(i_)       # (5.2s)

        # (6s) Activate candidate
        g_ = layer.fc['g_'][:, s] = (
            np.dot(X, layer.p['Ug'])
            + np.dot(hp, layer.p['Vg'])
            + layer.p['bg']
        )   # (6.1s)

        g = layer.fc['g'][:, s] = layer.activate_candidate(g_)   # (6.2s)

        # (7s) Activate output gate
        o_ = layer.fc['o_'][:, s] = (
            np.dot(X, layer.p['Uo'])
            + np.dot(hp, layer.p['Vo'])
            + layer.p['bo']
        )   # (7.1s)

        o = layer.fc['o'][:, s] = layer.activate_output(o_)      # (7.2s)

        # (8s) Compute current memory state
        C_ = layer.fc['C_'][:, s] = (
            Cp_ * f
            + i * g
        )   # (8.1s)

        C = layer.fc['C'][:, s] = layer.activate(C_)             # (8.2s)

        # (9s) Compute current hidden state
        h = layer.fc['h'][:, s] = o * C

    # Return the last hidden state or the full sequence
    A = layer.fc['h'] if layer.sequences else layer.fc['h'][:, -1]

    return A    # To next layer



# EpyNN/epynn/embedding/parameters.py
# Related third party imports


def embedding_compute_shapes(layer, A):
    """Compute forward shapes and dimensions from input for layer.
    """
    X = A    # Input of current layer

    layer.fs['X'] = X.shape    #  (m, .. )

    layer.d['m'] = layer.fs['X'][0]        # Number of samples (m)
    layer.d['n'] = X.size // layer.d['m']  # Number of features (n)

    return None


def embedding_initialize_parameters(layer):
    """Initialize parameters from shapes for layer.
    """
    # No parameters to initialize for Embedding layer

    return None


def embedding_compute_gradients(layer):
    """Compute gradients with respect to weight and bias for layer.
    """
    # No gradients to compute for Embedding layer

    return None


def embedding_update_parameters(layer):
    """Update parameters from gradients for layer.
    """
    # No parameters to update for Embedding layer

    return None



# EpyNN/epynn/template/parameters
# Local application/library specific imports
from epynn.commons.models import Layer
from epynn.commons.maths import activation_tune
from epynn.template.forward import template_forward
from epynn.template.backward import template_backward
from epynn.template.parameters import (
    template_compute_shapes,
    template_initialize_parameters,
    template_compute_gradients,
    template_update_parameters
)


class Template(Layer):
    """
    Definition of a template layer prototype. This is a pass-through or inactive layer prototype which contains method definitions used for all active layers. For all layer prototypes, methods are wrappers of functions which contain the specific implementations.
    """

    def __init__(self):
        """Initialize instance variable attributes. Extended with ``super().__init__()`` which calls :func:`epynn.commons.models.Layer.__init__()` defined in the parent class.

        :ivar trainable: Whether layer's parameters should be trainable.
        :vartype trainable: bool
        """
        super().__init__()

        self.trainable = True

        return None

    def compute_shapes(self, A):
        """Is a wrapper for :func:`epynn.template.parameters.template_compute_shapes()`.

        :param A: Output of forward propagation from *previous* layer.
        :type A: :class:`numpy.ndarray`
        """
        template_compute_shapes(self, A)

        return None

    def initialize_parameters(self):
        """Is a wrapper for :func:`epynn.template.parameters.template_initialize_parameters()`.
        """
        template_initialize_parameters(self)

        return None

    def forward(self, A):
        """Is a wrapper for :func:`epynn.template.forward.template_forward()`.

        :param A: Output of forward propagation from *previous* layer.
        :type A: :class:`numpy.ndarray`

        :return: Output of forward propagation for **current** layer.
        :rtype: :class:`numpy.ndarray`
        """
        self.compute_shapes(A)
        activation_tune(self.se_hPars)
        A = template_forward(self, A)
        self.update_shapes(self.fc, self.fs)

        return A

    def backward(self, dX):
        """Is a wrapper for :func:`epynn.template.backward.template_backward()`.

        :param dX: Output of backward propagation from *next* layer.
        :type dX: :class:`numpy.ndarray`

        :return: Output of backward propagation for **current** layer.
        :rtype: :class:`numpy.ndarray`
        """
        activation_tune(self.se_hPars)
        dX = template_backward(self, dX)
        self.update_shapes(self.bc, self.bs)

        return dX

    def compute_gradients(self):
        """Is a wrapper for :func:`epynn.template.parameters.template_compute_gradients()`. Dummy method, there are no gradients to compute in layer.
        """
        template_compute_gradients(self)

        return None

    def update_parameters(self):
        """Is a wrapper for :func:`epynn.template.parameters.template_update_parameters()`. Dummy method, there are no parameters to update in layer.
        """
        if self.trainable:
            template_update_parameters(self)

        return None



# EpyNN/epynn/dense/forward.py
# Related third party imports
import numpy as np


def initialize_forward(layer, A):
    """Forward cache initialization.

    :param layer: An instance of dense layer.
    :type layer: :class:`epynn.dense.models.Dense`

    :param A: Output of forward propagation from previous layer.
    :type A: :class:`numpy.ndarray`

    :return: Input of forward propagation for current layer.
    :rtype: :class:`numpy.ndarray`
    """
    X = layer.fc['X'] = A

    return X


def dense_forward(layer, A):
    """Forward propagate signal to next layer.
    """
    # (1) Initialize cache
    X = initialize_forward(layer, A)

    # (2) Linear activation X -> Z
    Z = layer.fc['Z'] = (
        np.dot(X, layer.p['W'])
        + layer.p['b']
    )   # This is the weighted sum

    # (3) Non-linear activation Z -> A
    A = layer.fc['A'] = layer.activate(Z)

    return A    # To next layer



# EpyNN/epynn/commons/io.py
# Related third party imports
import numpy as np

 
def index_elements_auto(X_data):
    """Determine elements size and generate dictionary for one-hot encoding or features or label.

    :param X_data: Dataset containing samples features or samples label.
    :type X_data: :class:`numpy.ndarray`

    :return: One-hot encoding converter.
    :rtype: dict[str or int or float, int]

    :return: One-hot decoding converter.
    :rtype: dict[int, str or int or float]

    :return: Vocabulary size.
    :rtype: int
    """
    X_data = X_data.flatten().tolist()       # All elements in 1D list

    elements = sorted(list(set(X_data)))     # Unique elements list
    elements_size = len(elements)            # Number of elements

    # Converters to encode and decode sequences
    element_to_idx = {w: i for i, w in enumerate(elements)}
    idx_to_element = {i: w for w, i in element_to_idx.items()}

    return element_to_idx, idx_to_element, elements_size


def scale_features(X_data):
    """Scale input array within [0, 1].

    :param X_data: Raw data.
    :type X_data: :class:`numpy.ndarray`

    :return: Normalized data.
    :rtype: :class:`numpy.ndarray`
    """
    X_data = (X_data-np.min(X_data)) / (np.max(X_data)-np.min(X_data))

    return X_data


def one_hot_encode(i, elements_size):
    """Generate one-hot encoding array.

    :param i: One-hot index for current word.
    :type i: int

    :param elements_size: Number of keys in the word to index encoder.
    :type elements_size: int

    :return: One-hot encoding array for current word.
    :rtype: :class:`numpy.ndarray`
    """
    one_hot = np.zeros(elements_size)

    one_hot[i] = 1.0    # Set 1 at index assigned to word

    return one_hot


def one_hot_encode_sequence(sequence, element_to_idx, elements_size):
    """One-hot encode sequence.

    :param sequence: Sequential data.
    :type sequence: list or :class:`numpy.ndarray`

    :param element_to_idx: Converter with word as key and index as value.
    :type element_to_idx: dict[str or int or float, int]

    :param elements_size: Number of keys in converter.
    :type elements_size: int

    :return: One-hot encoded sequence.
    :rtype: :class:`numpy.ndarray`
    """
    encoding = np.array([one_hot_encode(element_to_idx[word], elements_size) for word in sequence])

    return encoding


def one_hot_decode_sequence(sequence, idx_to_element):
    """One-hot decode sequence.

    :param sequence: One-hot encoded sequence.
    :type sequence: list or :class:`numpy.ndarray`

    :param idx_to_element: Converter with index as key and word as value.
    :type idx_to_element: dict[int, str or int or float]

    :return: One-hot decoded sequence.
    :rtype: list[str or int or float]
    """
    decoding = [idx_to_element[np.argmax(encoded)] for encoded in sequence]

    return decoding


def encode_dataset(X_data, element_to_idx, elements_size):
    """One-hot encode a set of sequences.

    :param X_data: Contains sequences.
    :type X_data: :class:`numpy.ndarray`

    :param element_to_idx: Converter with word as key and index as value.
    :type element_to_idx: dict[str or int or float, int]

    :param elements_size: Number of keys in converter.
    :type elements_size: int

    :return: One-hot encoded dataset.
    :rtype: list[:class:`numpy.ndarray`]
    """
    X_encoded = []

    # Iterate over sequences
    for i in range(X_data.shape[0]):

        sequence = X_data[i]    # Retrieve sequence

        encoded_sequence = one_hot_encode_sequence(sequence, element_to_idx, elements_size)

        X_encoded.append(encoded_sequence)    # Append to dataset of encoded sequences

    return X_encoded


def padding(X_data, padding, forward=True):
    """Image padding.

    :param X_data: Array representing a set of images.
    :type X_data: :class:`numpy.ndarray`

    :param padding: Number of zeros to add in each side of the image.
    :type padding: int

    :param forward: Set to False to remove padding, defaults to `True`.
    :type forward: bool, optional
    """
    if padding and forward:
        # Pad image
        shape = ((0, 0), (padding, padding), (padding, padding), (0, 0))
        X_data = np.pad(X_data, shape, mode='constant', constant_values=(0, 0))

    elif padding and not forward:
        # Remove padding
        X_data = X_data[:, padding:-padding, padding:-padding, :]

    return X_data



# EpyNN/epynn/network/report.py
# Standard library imports
import time

# Related third party imports
from tabulate import tabulate
from termcolor import cprint

# Local application/library specific imports
from epynn.network.evaluate import batch_evaluate
from epynn.commons.logs import (
    current_logs,
    dsets_samples_logs,
    dsets_labels_logs,
    headers_logs,
    initialize_logs_print,
    layers_lrate_logs,
    layers_others_logs,
    network_logs,
    start_counter
)


def model_report(model):
    """Report selected metrics for datasets at current epoch.

    :param model: An instance of EpyNN network object.
    :type model: :class:`epynn.network.models.EpyNN`
    """
    # You may edit the colorscheme to fulfill your preference
    colors = [
        'white',
        'green',
        'red',
        'magenta',
        'cyan',
        'yellow',
        'blue',
        'grey',
    ]

    # Rows in tabular report excluding headers
    size_table = 11

    # Initialize list of rows with headers
    if model.e == 0 or not hasattr(model, 'current_logs'):
        model.current_logs = [headers_logs(model, colors)]

    # Check if last epoch
    eLast = (model.e == model.epochs - 1)

    # Append row one every verboseth epoch or if last epoch
    if model.e % model.verbose == 0 or eLast:
        model.current_logs.append(current_logs(model, colors))

    # Report on terminal
    if len(model.current_logs) == size_table + 1 or eLast:

        logs = tabulate(model.current_logs,
                        headers="firstrow",
                        numalign="center",
                        stralign='center',
                        tablefmt="pretty",
                        )

        print('\n')
        print (logs, flush=True)

        # Clear-up
        del model.current_logs

    return None


def single_batch_report(model, batch, A):
    """Report accuracy and cost for current batch.

    :param model: An instance of EpyNN network.
    :type model: :class:`epynn.network.models.EpyNN`

    :param batch: An instance of batch dataSet.
    :type batch: :class:`epynn.commons.models.dataSet`

    :param A: Output of forward propagation for batch.
    :type A: :class:`numpy.ndarray`
    """
    current = time.time()

    # Total elapsed time
    elapsed_time = round(current - model.ts, 2)

    # Time for one epoch based on current batch
    epoch_time = (current - model.cts) * len(model.embedding.batch_dtrain)
    model.cts = current

    # Epochs per second
    rate = round((model.e + 1) / (elapsed_time + 1e-16), 3)

    # Time until completion
    ttc = round((model.epochs - model.e + 1) / (rate + 1e-16))

    # Accuracy and cost
    accuracy, cost = batch_evaluate(model, batch.Y, A)
    accuracy = round(accuracy, 3)
    cost = round(cost, 5)

    # Current batch numerical identifier
    batch_counter = batch.name + '/' + model.embedding.batch_dtrain[-1].name

    # Format and print data
    rate = '{:.2e}'.format(rate)

    log = ('Epoch %s - Batch %s - Accuracy: %s Cost: %s - TIME: %ss RATE: %se/s TTC: %ss'
           % (model.e, batch_counter, accuracy, cost, elapsed_time, rate, ttc))

    cprint('{: <100}'.format(log), 'white', attrs=['bold'], end='\r', flush=True)

    return None


def initialize_model_report(model, timeout):
    """Report exhaustive initialization logs for datasets,
    model architecture and shapes, layers hyperparameters.

    :param model: An instance of EpyNN network.
    :type model: :class:`epynn.network.models.EpyNN`

    :param timeout: Time to hold on initialization logs.
    :type timeout: int
    """
    model.init_logs = []

    # Dataset initialization logs
    dsets = model.embedding.dsets
    se_dataset = model.embedding.se_dataset

    model.init_logs.append(dsets_samples_logs(dsets, se_dataset))
    model.init_logs.append(dsets_labels_logs(dsets))

    # Model architecture and shapes initialization logs
    network = model.network

    model.init_logs.append(network_logs(network))

    # Model and layer hyperparameters initialization logs
    layers = model.layers

    model.init_logs.append(layers_lrate_logs(layers))
    model.init_logs.append(layers_others_logs(layers))

    initialize_logs_print(model)

    start_counter(timeout)

    return None



# EpyNN/epynnlive/dummy_time/prepare_dataset.py
# Standard library imports
import random

# Related third party imports
import numpy as np


def features_time(TIME=1, SAMPLING_RATE=128):
    """Generate dummy time features.

    Time features may be white noise or a sum with a pure sine-wave.

    The pure sin-wave has random frequency lower than SAMPLING_RATE // 4.

    :param SAMPLING_RATE: Sampling rate (Hz), defaults to 128.
    :type SAMPLING_RATE: int

    :param TIME: Sampling time (s), defaults to 1.
    :type TIME: int

    :return: Time features of shape (N_FEATURES,).
    :rtype: :class:`numpy.ndarray`

    :return: White noise of shape (N_FEATURES,).
    :rtype: :class:`numpy.ndarray`
    """
    # Number of features describing a sample
    N_FEATURES = TIME * SAMPLING_RATE

    # Initialize features array
    features = np.linspace(0, TIME, N_FEATURES, endpoint=False)

    # Random choice of true signal frequency
    signal_frequency = random.uniform(0, SAMPLING_RATE // 4)

    # Generate pure sine wave of N_FEATURES points
    features = np.sin(2 * np.pi * signal_frequency * features)

    # Generate white noise
    white_noise = np.random.normal(0, scale=0.5, size=N_FEATURES)

    # Random choice between noisy signal or white noise
    features = random.choice([features + white_noise, white_noise])

    return features, white_noise


def label_features(features, white_noise):
    """Prepare label associated with features.

    The dummy law is:

    True signal in features = positive.
    No true signal in features = negative.

    :return: Time features of shape (N_FEATURES,).
    :rtype: :class:`numpy.ndarray`

    :return: White noise of shape (N_FEATURES,).
    :rtype: :class:`numpy.ndarray`

    :return: Single-digit label with respect to features.
    :rtype: int
    """
    # Single-digit positive and negative labels
    p_label = 0
    n_label = 1

    # Test if features contains signal (0)
    if any(features != white_noise):
        label = p_label

    # Test if features is equal to white noise (1)
    elif all(features == white_noise):
        label = n_label

    return label


def prepare_dataset(N_SAMPLES=100):
    """Prepare a set of dummy time sample features and label.

    :param N_SAMPLES: Number of samples to generate, defaults to 100.
    :type N_SAMPLES: int

    :return: Set of sample features.
    :rtype: tuple[:class:`numpy.ndarray`]

    :return: Set of single-digit sample label.
    :rtype: tuple[int]
    """
    # Initialize X and Y datasets
    X_features = []
    Y_label = []

   # Iterate over N_SAMPLES
    for i in range(N_SAMPLES):

        # Compute random time features
        features, white_noise = features_time()

        # Retrieve label associated with features
        label = label_features(features, white_noise)

        # From n measurements to n steps with 1 measurements
        features = np.expand_dims(features, 1)

        # Append sample features to X_features
        X_features.append(features)

        # Append sample label to Y_label
        Y_label.append(label)

    # Prepare X-Y pairwise dataset
    dataset = list(zip(X_features, Y_label))

    # Shuffle dataset
    random.shuffle(dataset)

    # Separate X-Y pairs
    X_features, Y_label = zip(*dataset)

    return X_features, Y_label



# EpyNN/epynn/rnn/parameters.py
# Related third party imports
import numpy as np


def rnn_compute_shapes(layer, A):
    """Compute forward shapes and dimensions from input for layer.
    """
    X = A    # Input of current layer

    layer.fs['X'] = X.shape    # (m, s, e)

    layer.d['m'] = layer.fs['X'][0]    # Number of samples (m)
    layer.d['s'] = layer.fs['X'][1]    # Steps in sequence (s)
    layer.d['e'] = layer.fs['X'][2]    # Elements per step (e)

    # Shapes for trainable parameters         Unit cells (u)
    layer.fs['U'] = (layer.d['e'], layer.d['u'])    # (e, u)
    layer.fs['V'] = (layer.d['u'], layer.d['u'])    # (u, u)
    layer.fs['b'] = (1, layer.d['u'])               # (1, u)

    # Shape of hidden state (h) with respect to steps (s)
    layer.fs['h'] = (layer.d['m'], layer.d['s'], layer.d['u'])

    return None


def rnn_initialize_parameters(layer):
    """Initialize trainable parameters from shapes for layer.
    """
    # For linear activation of hidden state (h_)
    layer.p['U'] = layer.initialization(layer.fs['U'], rng=layer.np_rng)
    layer.p['V'] = layer.initialization(layer.fs['V'], rng=layer.np_rng)
    layer.p['b'] = np.zeros(layer.fs['b']) # dot(X, U) + dot(hp, V) + b

    return None


def rnn_compute_gradients(layer):
    """Compute gradients with respect to weight and bias for layer.
    """
    # Gradients initialization with respect to parameters
    for parameter in layer.p.keys():
        gradient = 'd' + parameter
        layer.g[gradient] = np.zeros_like(layer.p[parameter])

    # Reverse iteration over sequence steps
    for s in reversed(range(layer.d['s'])):

        dh_ = layer.bc['dh_'][:, s]  # Gradient w.r.t hidden state h_
        X = layer.fc['X'][:, s]      # Input for current step
        hp = layer.fc['hp'][:, s]    # Previous hidden state

        # (1) Gradients of the loss with respect to U, V, b
        layer.g['dU'] += np.dot(X.T, dh_)     # (1.1) dL/dU
        layer.g['dV'] += np.dot(hp.T, dh_)    # (1.2) dL/dV
        layer.g['db'] += np.sum(dh_, axis=0)  # (1.3) dL/db

    return None


def rnn_update_parameters(layer):
    """Update parameters from gradients for layer.
    """
    for gradient in layer.g.keys():
        parameter = gradient[1:]
        # Update is driven by learning rate and gradients
        layer.p[parameter] -= layer.lrate[layer.e] * layer.g[gradient]

    return None



# EpyNN/epynnlive/author_music/prepare_dataset.py
# Standard library imports
import tarfile
import random
import glob
import os

# Related third party imports
import wget
import numpy as np
from scipy.io import wavfile

# Local application/library specific imports
from epynn.commons.logs import process_logs


def download_music():
    """Download some guitar music.
    """
    data_path = os.path.join('.', 'data')

    if not os.path.exists(data_path):

        # Download @url with wget
        url = 'https://synthase.s3.us-west-2.amazonaws.com/author_music.tar'
        fname = wget.download(url)

        # Extract archive
        tar = tarfile.open(fname).extractall('.')
        process_logs('Make: '+fname, level=1)

        # Clean-up
        os.remove(fname)

    return None


def clips_music(wav_file, TIME=1, SAMPLING_RATE=10000):
    """Clip music and proceed with resampling.

    :param wav_file: The filename of .wav file which contains the music.
    :type wav_file: str

    :param SAMPLING_RATE: Sampling rate (Hz), default to 10000.
    :type SAMPLING_RATE: int

    :param TIME: Sampling time (s), defaults to 1.
    :type TIME: int

    :return: Clipped and re-sampled music.
    :rtype: list[:class:`numpy.ndarray`]
    """
    # Number of features describing a sample
    N_FEATURES = int(SAMPLING_RATE * TIME)

    # Retrieve original sampling rate (Hz) and data
    wav_sampling_rate, wav_data = wavfile.read(wav_file)

    # 16-bits wav files - Pass all positive and norm. [0, 1]
    # wav_data = (wav_data + 32768.0) / (32768.0 * 2)
    wav_data = wav_data.astype('int64')
    wav_data = (wav_data + np.abs(np.min(wav_data)))
    wav_data = wav_data / np.max(wav_data)

    # Digitize in 4-bits signal
    n_bins = 16
    bins = [i / (n_bins - 1) for i in range(n_bins)]
    wav_data = np.digitize(wav_data, bins, right=True)

    # Compute step for re-sampling
    sampling_step = int(wav_sampling_rate / SAMPLING_RATE)

    # Re-sampling to avoid memory allocation errors
    wav_resampled = wav_data[::sampling_step]

    # Total duration (s) of original data
    wav_time = wav_data.shape[0] / wav_sampling_rate

    # Number of clips to slice from original data
    N_CLIPS = int(wav_time / TIME)

    # Make clips from data
    clips = [wav_resampled[i * N_FEATURES:(i+1) * N_FEATURES] for i in range(N_CLIPS)]

    return clips


def prepare_dataset(N_SAMPLES=100):
    """Prepare a dataset of clipped music as NumPy arrays.

    :param N_SAMPLES: Number of clip samples to retrieve, defaults to 100.
    :type N_SAMPLES: int

    :return: Set of sample features.
    :rtype: tuple[:class:`numpy.ndarray`]

    :return: Set of single-digit sample label.
    :rtype: tuple[:class:`numpy.ndarray`]
    """
    # Initialize X and Y datasets
    X_features = []
    Y_label = []

    wav_paths = os.path.join('data', '*', '*wav')

    wav_files = glob.glob(wav_paths)

    # Iterate over WAV_FILES
    for wav_file in wav_files:

        # Retrieve clips
        clips = clips_music(wav_file)

        # Iterate over clips
        for features in clips:

            # Clip is positive if played by true author (0) else negative (1)
            label = 0 if 'true' in wav_file else 1

            # Append sample features to X_features
            X_features.append(features)

            # Append sample label to Y_label
            Y_label.append(label)

    # Prepare X-Y pairwise dataset
    dataset = list(zip(X_features, Y_label))

    # Shuffle dataset
    random.shuffle(dataset)

    # Truncate dataset to N_SAMPLES
    dataset = dataset[:N_SAMPLES]

    # Separate X-Y pairs
    X_features, Y_label = zip(*dataset)

    return X_features, Y_label



# EpyNN/epynn/convolution/backward.py
# Related third party imports
import numpy as np


def initialize_backward(layer, dX):
    """Backward cache initialization.

    :param layer: An instance of convolution layer.
    :type layer: :class:`epynn.convolution.models.Convolution`

    :param dX: Output of backward propagation from next layer.
    :type dX: :class:`numpy.ndarray`

    :return: Input of backward propagation for current layer.
    :rtype: :class:`numpy.ndarray`
    """
    dA = layer.bc['dA'] = dX

    return dA


def convolution_backward(layer, dX):
    """Backward propagate error gradients to previous layer.
    """
    # (1) Initialize cache
    dA = initialize_backward(layer, dX)    # (m, oh, ow, u)

    # (2) Gradient of the loss w.r.t. Z
    dZ = layer.bc['dZ'] = (
        dA
        * layer.activate(layer.fc['Z'], deriv=True)
    )   # dL/dZ

    # (3) Restore kernel dimensions
    dZb = dZ
    dZb = np.expand_dims(dZb, axis=3)
    dZb = np.expand_dims(dZb, axis=3)
    dZb = np.expand_dims(dZb, axis=3)
    # (m, oh, ow, 1, u)         ->
    # (m, oh, ow, 1, 1, u)     ->
    # (m, oh, ow, 1, 1, 1, u)

    # (4) Initialize backward output dL/dX
    dX = np.zeros_like(layer.fc['X'])      # (m, h, w, d)

    # Iterate over forward output height
    for oh in range(layer.d['oh']):

        hs = oh * layer.d['sh']
        he = hs + layer.d['fh']

        # Iterate over forward output width
        for ow in range(layer.d['ow']):

            ws = ow * layer.d['sw']
            we = ws + layer.d['fw']

            # (5hw) Gradient of the loss w.r.t Xb
            dXb = dZb[:, oh, ow, :] * layer.p['W']
            # (m, oh, ow,  1,  1, 1, u) - dZb
            #         (m,  1,  1, 1, u) - dZb[:, h, w, :]
            #            (fh, fw, d, u) - W

            # (6hw) Sum over units axis
            dX[:, hs:he, ws:we, :] += np.sum(dXb, axis=4)
            # (m, fh, fw, d, u) - dXb
            # (m, fh, fw, d)    - np.sum(dXb, axis=4)

    layer.bc['dX'] = dX

    return dX



# EpyNN/epynn/dropout/forward.py


def initialize_forward(layer, A):
    """Forward cache initialization.

    :param layer: An instance of dropout layer.
    :type layer: :class:`epynn.dropout.models.Dropout`

    :param A: Output of forward propagation from previous layer.
    :type A: :class:`numpy.ndarray`

    :return: Input of forward propagation for current layer.
    :rtype: :class:`numpy.ndarray`
    """
    X = layer.fc['X'] = A

    return X

import numpy as np
def dropout_forward(layer, A):
    """Forward propagate signal to next layer.
    """
    # (1) Initialize cache
    X = initialize_forward(layer, A)

    # (2) Generate dropout mask
    D = layer.np_rng.uniform(0, 1, layer.fs['D'])

    # (3) Apply a step function with respect to drop_prob (k)
    D = layer.fc['D'] = (D > layer.d['d'])

    # (4) Drop data points
    A = X * D

    # (5) Scale up signal
    A /= (1 - layer.d['d'])

    return A    # To next layer



# EpyNN/epynn/flatten/backward.py
# Related third party imports
import numpy as np


def initialize_backward(layer, dX):
    """Backward cache initialization.

    :param layer: An instance of flatten layer.
    :type layer: :class:`epynn.flatten.models.Flatten`

    :param dX: Output of backward propagation from next layer.
    :type dX: :class:`numpy.ndarray`

    :return: Input of backward propagation for current layer.
    :rtype: :class:`numpy.ndarray`
    """
    dA = layer.bc['dA'] = dX

    return dA


def flatten_backward(layer, dX):
    """Backward propagate error gradients to previous layer.
    """
    # (1)
    dA = initialize_backward(layer, dX)

    # (2) Reverse reshape (m, n) -> (m, ...)
    dX = layer.bc['dX'] = np.reshape(dA, layer.fs['X'])

    return dX    # To previous layer



# EpyNN/epynn/initialize.py
# Local application/library specific imports
from epynn.commons.logs import set_highlighted_excepthook
from epynn.commons.library import settings_verification

# Import epynn.settings in working directory if not present
settings_verification()

# Colored excepthook
set_highlighted_excepthook()



# EpyNN/epynn/embedding/parameters.py
# Related third party imports
import numpy as np

# Local application/library specific imports
from epynn.commons.models import Layer
from epynn.embedding.dataset import (
    embedding_prepare,
    embedding_encode,
    embedding_check,
    mini_batches,
)
from epynn.embedding.forward import embedding_forward
from epynn.embedding.backward import embedding_backward
from epynn.embedding.parameters import (
    embedding_compute_shapes,
    embedding_initialize_parameters,
    embedding_compute_gradients,
    embedding_update_parameters
)


class Embedding(Layer):
    """
    Definition of an embedding layer prototype.

    :param X_data: Dataset containing samples features, defaults to `None` which returns an empty layer.
    :type X_data: list[list[float or str or list[float or str]]] or NoneType, optional

    :param Y_data: Dataset containing samples label, defaults to `None`.
    :type Y_data: list[int or list[int]] or NoneType, optional

    :param relative_size: For training, validation and testing sets. Defaults to `(2, 1, 1)`
    :type relative_size: tuple[int], optional

    :param batch_size: For training batches, defaults to None which makes a single batch out of the training data.
    :type batch_size: int or NoneType, optional

    :param X_encode: Set to True to one-hot encode features, default to `False`.
    :type encode: bool, optional

    :param Y_encode: Set to True to one-hot encode labels, default to `False`.
    :type encode: bool, optional

    :param X_scale: Normalize sample features within [0, 1], default to `False`.
    :type X_scale: bool, optional
    """

    def __init__(self,
                 X_data=None,
                 Y_data=None,
                 relative_size=(2, 1, 0),
                 batch_size=None,
                 X_encode=False,
                 Y_encode=False,
                 X_scale=False):
        """Initialize instance variable attributes.
        """
        super().__init__()

        self.se_dataset = {
            'dtrain_relative': relative_size[0],
            'dval_relative': relative_size[1],
            'dtest_relative': relative_size[2],
            'batch_size': batch_size,
            'X_scale': X_scale,
            'X_encode': X_encode,
            'Y_encode': Y_encode,
        }

        X_data, Y_data = embedding_check(X_data, Y_data, X_scale)

        X_data, Y_data = embedding_encode(self, X_data, Y_data, X_encode, Y_encode)

        embedded_data = embedding_prepare(self, X_data, Y_data)

        self.dtrain, self.dval, self.dtest = embedded_data

        # Keep non-empty datasets
        self.dsets = [self.dtrain, self.dval, self.dtest]
        self.dsets = [dset for dset in self.dsets if dset.active]

        self.trainable = False

        return None

    def training_batches(self, init=False):
        """Wrapper for :func:`epynn.embedding.dataset.mini_batches()`.

        :param init: Wether to prepare a zip of X and Y data, defaults to False.
        :type init: bool, optional
        """
        if init:
            self.dtrain_zip = list(zip(self.dtrain.X, self.dtrain.Y))

        self.batch_dtrain = mini_batches(self)

        return None

    def compute_shapes(self, A):
        """Wrapper for :func:`epynn.embedding.parameters.embedding_compute_shapes()`.

        :param A: Output of forward propagation from previous layer.
        :type A: :class:`numpy.ndarray`
        """
        embedding_compute_shapes(self, A)

        return None

    def initialize_parameters(self):
        """Wrapper for :func:`epynn.embedding.parameters.embedding_initialize_parameters()`.
        """
        embedding_initialize_parameters(self)

        return None

    def forward(self, A):
        """Wrapper for :func:`epynn.embedding.forward.embedding_forward()`.

        :param A: Output of forward propagation from *previous* layer.
        :type A: :class:`numpy.ndarray`

        :return: Output of forward propagation for **current** layer.
        :rtype: :class:`numpy.ndarray`
        """
        A = embedding_forward(self, A)
        self.update_shapes(self.fc, self.fs)

        return A

    def backward(self, dX):
        """Wrapper for :func:`epynn.embedding.backward.embedding_backward()`.

        :param dX: Output of backward propagation from next layer.
        :type dX: :class:`numpy.ndarray`

        :return: Output of backward propagation for current layer.
        :rtype: :class:`numpy.ndarray`
        """
        dX = embedding_backward(self, dX)
        self.update_shapes(self.bc, self.bs)

        return dX

    def compute_gradients(self):
        """Wrapper for :func:`epynn.embedding.parameters.embedding_compute_gradients()`. Dummy method, there are no gradients to compute in layer.
        """
        embedding_compute_gradients(self)

        return None

    def update_parameters(self):
        """Wrapper for :func:`epynn.embedding.parameters.embedding_update_parameters()`. Dummy method, there are no parameters to update in layer.
        """
        if self.trainable:
            embedding_update_parameters(self)

        return None



# EpyNN/epynn/embedding/forward.py


def initialize_forward(layer, A):
    """Forward cache initialization.

    :param layer: An instance of embedding layer.
    :type layer: :class:`epynn.embedding.models.Embedding`

    :param A: Output of forward propagation from previous layer.
    :type A: :class:`numpy.ndarray`

    :return: Input of forward propagation for current layer.
    :rtype: :class:`numpy.ndarray`
    """
    X = layer.fc['X'] = A

    return X


def embedding_forward(layer, A):
    """Forward propagate signal to next layer.
    """
    # (1) Initialize cache
    X = initialize_forward(layer, A)

    # (2) Pass forward
    A = layer.fc['A'] = X

    return A   # To next layer
